﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Reflection;
using System.Text;
using Telerik.WinControls.Data;
using Telerik.WinControls.UI;

namespace VirtualRadGridViewSortingFiltering.Core
{
    public class VirtualGridTableElement : GridTableElement
    {
        protected override Type ThemeEffectiveType
        {
            get
            {
                return typeof(GridTableElement);
            }
        }

        public override bool EnsureCellVisible(GridViewRowInfo rowInfo, GridViewColumn column)
        {
            if (this.MasterTemplate.Templates.Count > 0)
            {
                UpdateLayout();
            }

            EnsureRowVisible(rowInfo);

            if (!this.IsInValidState(true))
            {
                return false;
            }
            
            if (!this.ElementTree.Control.Visible || this.ElementTree.Control.Size == Size.Empty)
            {
                return false;
            }

            this.EnsureRowVisible(rowInfo);

            GridRowElement row = this.GetRowElement(rowInfo);

            if (row == null || column == null || !column.IsVisible || column.IsGrouped)
            {
                return false;
            }

            if (column.OwnerTemplate.Parent == null && column.IsPinned)
            {
                return false;
            }

            Rectangle clientRect = Rectangle.Empty;

            if (rowInfo.PinPosition == PinnedRowPosition.Top)
            {
                clientRect = this.ViewElement.TopPinnedRows.ControlBoundingRectangle;
            }
            else if (rowInfo.PinPosition == PinnedRowPosition.Bottom)
            {
                clientRect = this.ViewElement.BottomPinnedRows.ControlBoundingRectangle;
            }
            else
            {
                clientRect = this.ViewElement.ScrollableRows.ControlBoundingRectangle;
            }

            bool isRightToLeft = row.ElementTree.Control.RightToLeft == System.Windows.Forms.RightToLeft.Yes;
            GridVirtualizedRowElement virtualizedRow = row as GridVirtualizedRowElement;

            if (virtualizedRow != null)
            {
                clientRect.X = virtualizedRow.ScrollableColumns.ControlBoundingRectangle.X;

                if (isRightToLeft)
                {
                    clientRect.X -= virtualizedRow.TableElement.CellSpacing;
                }

                clientRect.Width = virtualizedRow.ScrollableColumns.ControlBoundingRectangle.Width;
            }

            object helper = typeof(GridTableElement).GetField("visibilityHelper", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(this);
            MethodInfo changeScrollbarValueMi = helper.GetType().GetMethod("ChangeScrollbarValue", BindingFlags.Instance | BindingFlags.NonPublic,null, new Type[] { typeof(RadScrollBarElement), typeof(int) }, null);

            foreach (GridCellElement cell in row.VisualCells)
            {
                if (cell.ColumnInfo != null && cell.ColumnInfo.IsPinned)
                {
                    continue;
                }

                if (cell.ColumnInfo == column)
                {
                    int delta;

                    if (cell.ControlBoundingRectangle.Width > clientRect.Width)
                    {
                        return false;
                    }

                    if (cell.ControlBoundingRectangle.Right > clientRect.Right + 1)
                    {
                        if (this.ColumnScroller.ScrollMode == ItemScrollerScrollModes.Discrete)
                        {
                            int maximum = this.ColumnScroller.Scrollbar.Maximum - this.ColumnScroller.Scrollbar.LargeChange + 1;

                            int value = column.Index;

                            if (value > maximum)
                            {
                                value = maximum;
                            }

                            this.ColumnScroller.Scrollbar.Value = value;
                        }
                        else
                        {
                            delta = cell.ControlBoundingRectangle.Right - clientRect.Right;

                            if (isRightToLeft)
                            {
                                delta *= -1;
                            }

                            changeScrollbarValueMi.Invoke(helper, new object[] { this.ColumnScroller.Scrollbar, this.ColumnScroller.Scrollbar.Value + delta });
                        }
                    }

                    if (cell.ControlBoundingRectangle.X < clientRect.X - 1)
                    {
                        if (this.ColumnScroller.ScrollMode == ItemScrollerScrollModes.Discrete)
                        {
                            int maximum = this.ColumnScroller.Scrollbar.Maximum - this.ColumnScroller.Scrollbar.LargeChange + 1;

                            int value = column.Index;

                            if (value > maximum)
                            {
                                value = maximum;
                            }

                            this.ColumnScroller.Scrollbar.Value = value;
                        }
                        else
                        {
                            delta = cell.ControlBoundingRectangle.X - clientRect.X;

                            if (isRightToLeft)
                            {
                                delta *= -1;
                            }

                            changeScrollbarValueMi.Invoke(helper, new object[] { this.ColumnScroller.Scrollbar, this.ColumnScroller.Scrollbar.Value + delta });
                        }
                    }

                    return true;
                }
            }

            if (rowInfo is GridViewGroupRowInfo)
            {
                return false;
            }

            ColumnGroupRowLayout rowLayout = this.ViewElement.RowLayout as ColumnGroupRowLayout;

            if (rowLayout != null)
            {
                int offset = rowLayout.GetColumnOffset(column);
                changeScrollbarValueMi.Invoke(helper, new object[] { this.ColumnScroller.Scrollbar, offset });

                return true;
            }

            int oldValue = this.ColumnScroller.Scrollbar.Value;
            bool res = false;

            if (this.ColumnScroller.ScrollMode == ItemScrollerScrollModes.Discrete)
            {
                int maximum = this.ColumnScroller.Scrollbar.Maximum - this.ColumnScroller.Scrollbar.LargeChange + 1;

                int value = column.Index;

                if (value > maximum)
                {
                    value = maximum;
                }

                this.ColumnScroller.Scrollbar.Value = value;
                this.Invalidate();
                res = true;
            }
            else
            {
                res = this.ColumnScroller.ScrollToItem(column, false);
            }

            if (oldValue < this.ColumnScroller.Scrollbar.Value)
            {
                this.UpdateLayout();

                foreach (GridCellElement cell in row.VisualCells)
                {
                    if (cell.ColumnInfo != null && cell.ColumnInfo.IsPinned)
                    {
                        continue;
                    }

                    if (cell.ColumnInfo == column)
                    {
                        int delta = isRightToLeft ?
                            clientRect.Left - cell.ControlBoundingRectangle.Left :
                            cell.ControlBoundingRectangle.Right - clientRect.Right;
                        changeScrollbarValueMi.Invoke(helper, new object[] { this.ColumnScroller.Scrollbar, this.ColumnScroller.Scrollbar.Value + delta });

                    }
                }
            }

            return res;

        }

        public override bool EnsureRowVisible(GridViewRowInfo rowInfo)
        {
            if (rowInfo == null || !rowInfo.IsVisible || MasterTemplate == null )//|| (bool)this.MasterTemplate.GetType().GetProperty("SuspendEnsureVisible", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(this.MasterTemplate, new object[] { }))
            {
                return false;
            }

            if (!this.GridViewElement.UseScrollbarsInHierarchy && this.ViewInfo.ParentRow != null)
            {
                return this.GridViewElement.TableElement.EnsureRowVisible(rowInfo);
            }

            if (rowInfo.IsPinned && rowInfo.ViewInfo == this.ViewInfo)
            {
                return false;
            }

            if (MasterTemplate.Templates.Count > 0 && !GridViewElement.UseScrollbarsInHierarchy)
            {
                typeof(GridTableElement).GetMethod("EnsureRowVisibleCore", BindingFlags.Instance | BindingFlags.NonPublic).Invoke(this, new object[] { rowInfo });

                GridTableElement tableElement = this.GridViewElement.GetRowView(rowInfo.ViewInfo) as GridTableElement;

                if (tableElement != null)
                {
                    GridRowElement rowElement = tableElement.GetRowElement(rowInfo);

                    if (rowElement != null)
                    {
                        return true;
                    }
                }

                this.ScrollToRow(rowInfo);
                return true;
            }

            if (rowInfo.ViewInfo != this.ViewInfo ||
              (rowInfo.ViewInfo == this.MasterTemplate.MasterViewInfo && rowInfo.PinPosition != PinnedRowPosition.None) ||
              (!this.ElementTree.Control.Visible && !(this.ElementTree.Control is MultiColumnComboGridView)) ||
              this.ElementTree.Control.Size == Size.Empty)
            {
                return false;
            }

            if (this.Size == SizeF.Empty)
            {
                return false;
            }

            if (this.MasterTemplate != null && this.MasterTemplate.EnablePaging)
            {
                int pageIndex = this.MasterTemplate.PageIndex;

                if (this.MasterTemplate.EnableGrouping && this.MasterTemplate.GroupDescriptors.Count > 0)
                {
                    int topmostGroupIndex = -1;
                    IEnumerator enumerator = new PagedGroupedTraverser(((RadDataView<GridViewRowInfo>)this.MasterTemplate.DataView).GroupBuilder.Groups.GroupList);

                    while (enumerator.MoveNext())
                    {
                        GridViewGroupRowInfo groupRow = enumerator.Current as GridViewGroupRowInfo;

                        if (groupRow != null)
                        {
                            topmostGroupIndex++;
                        }

                        if (enumerator.Current == rowInfo)
                        {
                            break;
                        }
                    }

                    if (topmostGroupIndex >= 0)
                    {
                        pageIndex = topmostGroupIndex / this.MasterTemplate.PageSize;
                    }
                }
                else
                {
                    pageIndex = this.ViewTemplate.DataView.GetItemPage(rowInfo);
                }

                if (pageIndex != this.MasterTemplate.PageIndex && !(rowInfo is GridViewNewRowInfo))
                {
                    this.MasterTemplate.MoveToPage(pageIndex);
                    this.ViewElement.UpdateRows(true);

                    if (this.GridViewElement != null)
                    {
                        this.GridViewElement.PagingPanelElement.UpdateView();
                    }
                }
            }

            this.UpdateLayout();

            GridViewDetailsRowInfo detailsRowInfo = rowInfo as GridViewDetailsRowInfo;

            if (detailsRowInfo == null)
            {
                object helper = typeof(GridTableElement).GetField("visibilityHelper", BindingFlags.Instance | BindingFlags.NonPublic).GetValue(this);
                return (bool)helper.GetType().GetMethod("EnsureRowVisibleCore", BindingFlags.Instance | BindingFlags.NonPublic).Invoke(helper, new object[] { rowInfo });
            }

            return false;
        }
    }
}
